/*	dhv.c	1.0	05/23/86	*/

#include	"dhv.h"
#if	NDHV > 0
/*
 * DHV-11 driver
 */

/*
 * The DHV has a 256 character fifo on it's receiver for all the channels
 * on that DHV.  Normally you would interrupt on every character that is
 * typed (or worse yet... uucp'ed).  If the flag POLLING is defined; the
 * driver utilizes the fifo by polling the DHV rather than interrupting on
 * every character.
 *
 * This can cause "bursty" output; but I think it is better than interrupting
 * on every character.
 *
 * When in polling mode; we let the DHV handle ^S/^Q's because we may not
 * check the fifo for upto sometime (see dhv_timer()).  This has "bad"
 * effects when running at high baud rates (like 19.2k).  If someone
 * changes their start/stop characters to something other than ^S/^Q; they
 * are asking for what they get.
 */
/*# define POLLED		/* If you want to poll rather than interrupt */
# define GANDALF	/* If you have a Gandalf Pacx switch */

# include	"../h/param.h"
# include	"../h/conf.h"
# include	"../h/dir.h"
# include	"../h/user.h"
# include	"../h/proc.h"
# include	"../h/ioctl.h"
# include	"../h/tty.h"
# include	"../h/map.h"
# include	"../h/buf.h"
# include	"../h/vm.h"

# include	"../is68kdev/qbvar.h"
# include	"../is68kdev/dhvreg.h"

# include	"../h/bkmac.h"
# include	"../h/clist.h"
# include	"../h/file.h"
# include	"../h/uio.h"

/*
 * Definition of the driver for the auto-configuration routines.  You may
 * notice that there isn't any modem control defined here as in a standard
 * DH.  This is because the modem control is handled more like a DZ in a
 * DHV.
 */
int	dhv_probe(),	/* Routine to look for a DHV on the bus */
	dhv_slave(),	/* dummy slave routine. */
	dhv_attach(),	/* After we found a DHV, we want to attach it */
	dvrint(),	/* Handle reciever interrupts */
	dvxint() ;	/* Handle transmitter interrupts */

/*
 * Place to hold the device specfic information (one for each DHV)
 */
struct	qb_device	*dhv_info[NDHV] ;

/*
 * A list of the standard addresses you would expect to find a DHV.
 */
u_short		*dhv_std[] = {
		(u_short *)0x3fe010,	/* 017760020 */
		(u_short *)0x3fe020,	/* 017760040 */
		(u_short *)0x3fe030,	/* 017760060 */
		(u_short *)0x3fe040,	/* 017760100 */
		(u_short *)0x0,
} ;

/*
 * Definitions of the auto-configuration routines for ioconf.c
 */
struct	qb_driver	DHVdriver = {
			dhv_probe,
			dhv_slave,
			dhv_attach,
			dhv_std,
			"dhv",
			dhv_info,
			"DHV"
} ;

#define	DHV_LINES	8	/* number of lines per DHV */
#define dhv_board(dev)	(minor(dev) >> 3)

/*
 * Default values.  ISPEED should be 19.2k or 38.4k baud depending on which
 * clock you want to use (A or B).  It is too much of a pain to support both
 */
#define	ISPEED	EXTA		/* set to EXTB for 38.4 */
#define IFLAGS	(EVENP | ODDP)	/* Any parity... we don't care */

/*
 * Software carrier.  For each DHV, the software carrier is defined in the
 * configuration file for that machine.  The low order 8 bits are used to
 * tell us whether or not we want modem control on that line.  For instance
 * if the software carrier bits were set to 0x0b then we would ignore modem
 * control on lines 0, 1, and 3.  The rest of the lines DTR, CD, RTS are
 * valid.
 */
short	dhv_softCAR[NDHV] ;

/*
 * For each line we will need a tty control structure.
 */
struct	tty	dhv_tty[NDHV * DHV_LINES] ;

/*
 * Declare transmitter DMA buffer space for each line.
 */
int	dhv_buf[NDHV*DHV_LINES][CBSIZE] ;

/*
 *  Number of characters per DHV (input count).  Used to determine polling
 * speed.  Gets set by dvrint and cleared by the timer routine.
 */
int	dhv_chars[NDHV] ;

# ifdef POLLED
/*
 * Variable to let us know if the timer has started.  Get's set on first open.
 */
int	timer_started = 0 ;
# endif

/*
 * Mask of active DHV's.
 */
int	dhv_act ;

int	dhv_start(), ttrstrt() ;
extern	int	hz ;
extern struct user u;
extern	int	cvec ;

/*
 * Depending on which clock we are going to use the speeds map
 * differently.  EXTA => clock B, EXTA => clock A.
 */
short	dhv_speeds[] = {
# if	ISPEED == EXTA
/* B clock */
	/* Zero  baud */	0,
	/* 50    baud */	0,
	/* 75    baud */	1,
	/* 110   baud */	2,
	/* 134.5 baud */	3,
	/* 150   baud */	4,
	/* 200   baud */	0,
	/* 300   baud */	5,
	/* 600   baud */	6,
	/* 1.2k  baud */	7,
	/* 1.8k  baud */	8,
	/* 2.4k  baud */	10,
	/* 4.8k  baud */	11,
	/* 9.6k  baud */	13,
	/* 19.2k baud */	14,
	/* 38.4k baud */	0,
# else	/* ISPEED == EXTB */
/* A clock */
	/* Zero  baud */	0,
	/* 50    baud */	0,
	/* 75    baud */	0,
	/* 110   baud */	2,
	/* 134.5 baud */	3,
	/* 150   baud */	0,
	/* 200   baud */	0,
	/* 300   baud */	5,
	/* 600   baud */	6,
	/* 1.2k  baud */	7,
	/* 1.8k  baud */	8,
	/* 2.4k  baud */	10,
	/* 4.8k  baud */	11,
	/* 9.6k  baud */	13,
	/* 19.2k baud */	0,
	/* 38.4k baud */	15,
# endif	/* ISPEED */
} ;

/*
 * Routine for auto-configuration to force a dhv interrupt.
 * Set the speed to 9600 and cause a transmitter interrupt.  Actually, we
 * will send a 0x0 out on channel 0.
 */
/*ARGSUSED*/
dhv_probe( reg )
caddr_t	reg ;
{
	register struct dhv_device *dhv_addr ;

# ifdef lint
	dvrint(0) ; dvxint(0) ;
# endif

	dhv_addr = (struct dhv_device *)reg ;
	dhv_addr->csr = DHV_TXIE ;			/* Turn on interrupts */
	dhv_addr->un.txchar = DHV_TX_DATA_VALID ;	/* let er rip... */

	DELAY( 3000 ) ;			/* wait for it to interrupt */

	dhv_addr->csr = 0 ;		/* turn off interrupts */

	/*
	 * Unix expects cvec to be the receiver, not the transmitter so
	 * we subtract 4 from cvec to fake it.
	 */
	if( cvec && cvec != 0x100 ) cvec -=4 ;

	return( sizeof(struct dhv_device) ) ;
}

/* Dummy slave routine...  Actually I should just set the address
 * in DHVdriver to be zero, but some systems puke on this.
 */
dhv_slave()
{
	return(1);
}

/*
 * Routine called to attach a DHV to the system.
 */
dhv_attach(qi)
struct	qb_device	*qi ;
{
	/* Save the software carrier flags for future reference */
	dhv_softCAR[qi->qi_unit] = qi->qi_flags ;
	return ;
}

/*
 * Open a single DHV line.
 */
/*ARGSUSED*/
dhv_open( dev, flag )
dev_t	dev ;
int	flag ;
{
register int			unit ;	/* contains minor number */
register int			dhv ;	/* which dhv */
register struct	tty		*tp ;
register struct qb_device	*qi ;
register struct	dhv_device	*dhv_addr ;
int	s ;
int	dhv_timer() ;
	
	unit = minor(dev) ;
	dhv = dhv_board( unit ) ;
	qi = dhv_info[dhv] ;
	/*
	 * Return an error if the unit number is illegal or the dhv
	 * wasn't found by autoconf.
	 */
	if( (unit >= (NDHV*DHV_LINES)) || (qi == 0) || (qi->qi_alive == 0) ) {
		return( ENXIO ) ;
	}
	tp = (struct tty *)&dhv_tty[unit] ;
	dhv_addr = (struct dhv_device *)((qi->qi_mi)->qm_addr) ;

	/*
	 * Check to see if this line is being used exclusively.
	 * We have some special code here to determine if we are a legit
	 * child of init (ie getty).  This code should allow us to use a
	 * line both as incomming and outgoing (Should really fix init to
	 * do this for us).
	 */
	if( (tp->t_state&TS_XCLUDE) &&
	    ( (u.u_uid != 0) || (u.u_procp->p_flag&SLOGIN)) ) {
		return( EBUSY ) ;
	}

	/*
	 * Set up the controlling tty structure for this user.
	 */
	tp->t_addr = (caddr_t)dhv_addr ;	/* address of this DHV */
	tp->t_oproc = dhv_start ;		/* re-start output routine */
	tp->t_state |= TS_WOPEN ;		/* waiting to be opened*/
	tp->t_dev = dev ;			/* major/minor of this dev */

	/*
	 * Set up the defaults for this tty.
	 */
	if( (tp->t_state & TS_ISOPEN) == 0 ) {
		(void)ttychars(tp) ;
		if( tp->t_ispeed == 0 ) {
			tp->t_ispeed = ISPEED ;	/* force A or B clock */
			tp->t_ospeed = ISPEED ;
			tp->t_flags = IFLAGS ;
			tp->t_state |= TS_HUPCLS ;
		}
		(void)dhv_param(unit) ;
	}

	/*
	 * set interrupt mask.
	 */
	 s = spl5() ;

	 /*
	  * Check if the DHV is active.  If not, fire up interrupts.
	  */
	if ( (dhv_act&(1<<dhv)) == 0 ) {
		dhv_act |= (1<<dhv) ;		/* mark as active */
# ifdef POLLED
		dhv_chars[dhv] = 0 ;		/* zero input count */
# endif
	}
# ifdef POLLED
	if( timer_started == 0 ) {
		timer_started++ ;
		(void)dhv_timer() ;
	}

	/*
	 * Select a line.  Note that this will disable interrupts, so
	 * we have to enable them.
	 */
	dhv_addr->csr = (unit & 0x7) | DHV_TXIE ;
# else
	dhv_addr->csr = (unit & 0x7) | DHV_IE ;
# endif POLLED

	/* Enable the line */
	dhv_addr->lnctrl = DHV_ON | DHV_LINK_TYPE ;	/* receiver */
	dhv_addr->unn.tbuffad2 |= DHV_TX_ENA ;		/* transmitter */

# ifdef POLLED
	/*
	 * Check if we want the dhv to handle flow control.
	 *
	 * I should check tchars to make sure that they are ^S/^Q.  On
	 * a DHV you can't change the definition.  I will leave this for
	 * the next revision.
	 */
	if ( (tp->t_flags & RAW) == 0 ) {
		dhv_addr->lnctrl |= DHV_OAUTO ;
	}
# endif

	/*
	 * Check if carrier is already on, if so set the flags for tty.
	 */
	if( dhv_addr->stat & DHV_DCD )
		tp->t_state |= TS_CARR_ON ;


	/*
	 * Check and see if we want modem control for this line.  If not,
	 * just turn the carrier on and don't worry about it.
	 */
	 if( dhv_softCAR[dhv] & (1 << (unit & 0x7)) ) {
		tp->t_state |= TS_CARR_ON ;
		(void)splx(s) ;
		return((*linesw[tp->t_line].l_open)(dev,tp)) ;
	}

	/*
	 * Do modem control.  We watch for CD (pin 8).
	 */
# ifdef TTY_NODELAY
	 if( (flag&FNDELAY) == 0 )	/* Do we really want to wait for CD */
# endif
	while( (tp->t_state&TS_CARR_ON) == 0 ) {
		(void)sleep( (caddr_t)&(tp->t_rawq), TTIPRI ) ;
		if( (tp->t_state&TS_XCLUDE) && (u.u_procp->p_flag&SLOGIN) ) {
			(void)splx(s) ;
			return(EBUSY) ;
		}
	}

	(void)splx(s) ;
	return((*linesw[tp->t_line].l_open)(dev,tp)) ;
}

/*
 * Close a DHV11 line.
 */
/*ARGSUSED*/
dhv_close(dev, flag)
dev_t	dev ;
int	flag ;
{
register struct tty		*tp ;
register struct dhv_device	*dhv_addr  ;
register int			s ;
extern	 int			wakeup() ;

	tp = (struct tty *)&dhv_tty[minor(dev)] ;
	dhv_addr = (struct dhv_device *)tp->t_addr ;
	(void)(*linesw[tp->t_line].l_close)(tp) ;  /* Close the tty line. */

	/*
	 * Turn off DTR & RTS (pin 20 & pin 6) -> LOW.
	 */
	if( (tp->t_state&(TS_HUPCLS|TS_WOPEN)) || (tp->t_state&TS_ISOPEN)==0) {
		s=spl5() ;
# ifdef POLLED
		dhv_addr->csr = (minor(dev) & 0x7) | DHV_TXIE ;
# else
		dhv_addr->csr = (minor(dev) & 0x7) | DHV_IE ;
# endif POLLED
		dhv_addr->lnctrl = DHV_OFF ;	/* turn off the dhv */

		tp->t_state &= ~TS_CARR_ON ;
# ifdef NOMDM
		tp->t_flags &= ~NOMDM ;
# endif NOMDM
		(void)splx(s) ;
# ifdef GANDALF
		/*
		 * Kludge for our Gandalf Pacx switch.  When it get's loaded
		 * down we have to wait a long time before we allow this line
		 * to be used again.  This should really be done by init.
		 */
		(void)timeout(wakeup, (caddr_t)&tp->t_dev, hz<<1 ) ;
# else
		/*
		 * For a standard modem we want to wait about 600ms; we
		 * will wait about a 1/2 second.  That should be good enough.
		 */
		(void)timeout(wakeup, (caddr_t)&tp->t_dev, hz>>1 ) ;
# endif GANDALF
		(void)sleep( (caddr_t)&tp->t_dev, PZERO ) ;	/* zzzzzzz */
	}

	(void)ttyclose(tp) ;
	return ;
}

dhv_read(dev, uio)
dev_t	dev ;
struct	uio	uio ;
{
register struct tty	*tp ;

	tp=(struct tty *)&dhv_tty[minor(dev)] ;
	return( (*linesw[tp->t_line].l_read)(tp, uio) ) ;
}

dhv_write(dev, uio)
dev_t	dev ;
struct	uio	uio ;
{
register struct tty	*tp ;

	tp=(struct tty *)&dhv_tty[minor(dev)] ;
	return( (*linesw[tp->t_line].l_write)(tp, uio) ) ;
}

/*
 * DHV11 receiver interrupt.
 * If POLLED is set this only gets called by dhv_timer.  Otherwise we
 * interrupt on every bloody character (yuk).
 */
dvrint( dhv )
int	dhv ;
{
register struct qb_device	*qi ;
register struct dhv_device	*dhv_addr ;
register struct tty		*tp0 ;
register struct tty		*tp ;
register unsigned		c ;
int				overrun = 0 ;

	qi = dhv_info[dhv] ;
	if( qi == 0 || qi->qi_alive == 0 ) 
		return ;
	dhv_addr = (struct dhv_device *) ((qi->qi_mi)->qm_addr) ;
	tp0 = (struct tty *)&dhv_tty[dhv<<3] ;
	/*
	 * We will check the data valid bit in RBUF rather than checking
	 * read data available in the CSR.  If you read the CSR you could
	 * lose the state of the TX.ACTION bit, which could mean that we
	 * miss the termination of a DMA transfer in progress.
	 *
	 * We check RBUF bit 15...
	 */
	while( ((c=dhv_addr->un.rbuf)&DHV_DATA_VALID) != 0 ) {
# ifdef POLLED
		/*
		 * Bump the character count for this DHV.  This is used by
		 * the dhv_timer routine to determine how often to poll.
		 */
		 dhv_chars[dhv]++ ;
# endif POLLED

		 /*
		  * Get the line number from the rcr and use it to compute
		  * the tty structure for this line.  Why they use 4 bits
		  * for the line number is beyond me.  They changed everything
		  * else about a dhv ... why not this.
		  */
		tp = tp0 + ((c >> 8) & 0x7 ) ;

		/*
		 * if RBUF<14:12>==111 (MDM_INFO) then the next character in
		 * the FIFO is a modem status character.
		 */
# define MDM_INFO (DHV_PARITY_ERR | DHV_OVERRUN_ERR | DHV_FRAME_ERR)
		if( (c&MDM_INFO) == MDM_INFO) {
			/*
			 * Is this Diagnostic information.  There seems
			 * to be a misprint in the DHV manual.  I am not
			 * sure how to check for Diagnostic information
			 * from the DHV.  Anyways, I wouldn't once I got
			 * it all I could do is print it and keep on going.
			 *
			 *if ( c & DHV_C_DIAG ) {
				continue ;
			}*/
			/*
			 * Check for LMDMBUF bit set for carrier.
			 */
			if ((tp->t_state&TS_WOPEN) == 0 &&
			    (tp->t_flags&MDMBUF)) {
				/*
				 * If DCD went high, then start output.
				 */
				if( (c<<8) & DHV_DCD ) {
					tp->t_state &= ~TS_TTSTOP ;
					ttstart(tp) ;
				/*
				 * If DCD went low and we aren't already
				 * stopped, stop output.
				 */
				} else if( (tp->t_state&TS_TTSTOP) == 0 ) {
					tp->t_state |= TS_TTSTOP ;
					dhv_stop(tp, 0) ;
				}
			/*
			 * If the carrier dropped & we care sent a hangup
			 * signal to all the processes and continue them.
			 */
			} else if (((c<<8)&DHV_DCD) == 0 &&
				   (dhv_softCAR[dhv]&(1<<((c>>8)&0x7)))==0) {
				if( (tp->t_state&TS_WOPEN) == 0 &&
				    (tp->t_flags&NOHANG) == 0     ) {
					gsignal(tp->t_pgrp, SIGHUP) ;
					gsignal(tp->t_pgrp, SIGCONT) ;
# ifdef POLLED
					dhv_addr->csr = ((c>>8)& 0x7)|DHV_TXIE;
# else
					dhv_addr->csr = ((c>>8)& 0x7)|DHV_IE;
# endif
					dhv_addr->lnctrl = DHV_OFF ;
					(void)(*linesw[tp->t_line].l_close)(tp);
					ttyflush(tp, FREAD|FWRITE) ;
				}
				tp->t_state &= ~TS_CARR_ON ;
# ifdef NOMDM
				tp->t_flags &= ~NOMDM ;
# endif
			/*
			 * Carrier came up (pin 8).
			 */
			} else if( (tp->t_state&TS_CARR_ON) == 0 ) {
				tp->t_state |= TS_CARR_ON ;
				wakeup((caddr_t)&tp->t_rawq) ;
			}
			/*
 			 * We should also do something about RI, CTS, & DSR
			 * here...
			 */
			continue ;
		}

		/*
		 * Let's wakeup the sleepers.
		 */
		 if( (tp->t_state&TS_ISOPEN) == 0 ) {
			(void)wakeup((caddr_t)&tp->t_rawq) ;
			continue ;
		}

		/*
		 * Check for parity errors on input.  Just through away
		 * the character if there is a parity error.  We only
		 * check for this if the parity is set to ODD or EVEN.
		 */
		if( c&DHV_PARITY_ERR )
			if( (tp->t_flags&(EVENP|ODDP)) == EVENP ||
			    (tp->t_flags&(EVENP|ODDP)) == ODDP )
				continue ;

		/*
		 * Check for overrun errors.  We only want to do this once
		 * otherwise we will swap the console with overrun messages.
		 */
		if( (c&DHV_OVERRUN_ERR) && (overrun == 0 ) ) {
			overrun = 1 ;
			(void)printf( "dhv%d: silo overflow.\n", dhv ) ;
		}

		/*
		 * Check for framing errors (break).  Generate a NULL if
		 * we are in raw mode (for getty), or an interrupt character
		 * if we are in cooked/cbreak mode.
		 */
		if( c&DHV_FRAME_ERR )
			if(tp->t_flags&RAW)	c = 0 ;
			else			c = tp->t_intrc ;
		
# ifdef POLLED
		/*
		 * If we are polling, we wait too long to process a ^S/^Q
		 * sequence.  So, if we aren't in RAW mode, we will let
		 * the hardware do it and eat ^S & ^Q's.  See dhv_param.
		 *
		 * Note:  We & with 0xff rather than 0x7f.  The dhv only
		 * passes 7 bits when parity is set anyways.
		 */
		if( ((tp->t_flags&RAW) == 0) &&
		     (((c&0xff) == CTRL(S)) || ((c&0xff) == CTRL(Q))) )
			continue ;
# endif
# if NBK > 0 || Nbk > 0
		if(tp->t_line == NETLDISC) {
			c &= 0x7f ;
			BKINPUT(c, tp ) ;
		} else
# endif
			(void)(*linesw[tp->t_line].l_rint)(c, tp) ;
	}

	return ;
}

/*
 * Ioctl for DHV11.
 */
/*ARGSUSED*/
dhv_ioctl(dev, cmd, data, flag)
dev_t	dev ;
int	cmd ;
caddr_t	data ;
int	flag ;
{
register	int		unit ;
register struct tty		*tp ;
register struct dhv_device	*dhv_addr ;
register int			s ;
int				error ;

	unit=minor(dev) ;
	tp = (struct tty *)&dhv_tty[unit] ;
	error = (*linesw[tp->t_line].l_ioctl)(tp, cmd, data, flag) ;
	if (error >= 0 )
		return (error) ;
	
	error = ttioctl( tp, cmd, data, flag ) ;
	if (error >= 0 ) {
		if (cmd == TIOCSETP || cmd == TIOCSETN)
			(void)dhv_param(unit) ;
		return (error) ;
	}

	dhv_addr = (struct dhv_device *)tp->t_addr ;
	/*
	 * Block interrupts so parameters will be set before the line
	 * interrupts.
	 */
	s = spl5() ;
	/* Select the line */
# ifdef POLLED
	dhv_addr->csr = (unit & 07) | DHV_TXIE ;
# else
	dhv_addr->csr = (unit & 07) | DHV_IE ;
# endif


	switch(cmd) {

	case TIOCSBRK:	/* Turn on the Break bit */
		dhv_addr->lnctrl |= DHV_BREAK ;
		break ;				/* No pun intended. */
	
	case TIOCCBRK:	/* Turn of the Break bit */
		dhv_addr->lnctrl &= ~DHV_BREAK ;
		break ;				/* Pun intended. */
	
	case TIOCSDTR:	/* Turn on DTR and RTS */
		dhv_addr->lnctrl |= DHV_DTR|DHV_RTS ;
		break ;
	
	case TIOCCDTR:	/* Turn of DTR & RTS */
		dhv_addr->lnctrl &= ~(DHV_DTR|DHV_RTS) ;
		break ;
	
	default:
		return( ENOTTY ) ;
	}
	(void)splx(s) ;

	return(0) ;
}

/*
 * Set parameters from open or stty into the DHV11 hardware registers.
 */
dhv_param( unit )
register int	unit ;
{
register struct tty		*tp ;
register struct dhv_device	*dhv_addr ;
register int			lpar ;
int				s ;

	tp= (struct tty *) &dhv_tty[unit] ;
	dhv_addr = (struct dhv_device *)tp->t_addr ;
	/*
	 * Block interrupts so parameters will be set before the line
	 * interrupts.
	 */
	s = spl5() ;
	/* Select the line */
# ifdef POLLED
	dhv_addr->csr = (unit & 07) | DHV_TXIE ;
# else
	dhv_addr->csr = (unit & 07) | DHV_IE ;
# endif

	/* Hang up if they set the line speed to zero */
	if( (tp->t_ispeed) == 0 ) {
		tp->t_state |= TS_HUPCLS ;
		(void)splx(s) ;
		(void)dhv_close(unit, 0); /* close this file descriptor */
		return ;
	}

	/*
	 * We will set and mask the variable lpar till we get it to
	 * our liking.  Then we will set the line parameter register
	 * to lpar.
	 */
	
	lpar = ((dhv_speeds[tp->t_ospeed] << 12) |	/* set up the speeds */
		(dhv_speeds[tp->t_ispeed] << 8 )) ;

	if( (tp->t_ispeed) == B134)
		lpar |= DHV_6BITS | DHV_PARITY_ENABLE ;
		/*
		 * We should set half duplex here... but I don't think there
		 * is anyway of doing this on a DHV11.
		 */
	/*
	 * If we are in raw mode, transparent output mode, or we don't
	 * care about parity we want toset up the DHV to transmit all 8 bits.
	 */
	else if( (tp->t_flags&(RAW|LITOUT)) || (tp->t_flags&ANYP) == ANYP)
		lpar |= DHV_8BITS ;
	else
		/*
	 	* Turn on parity enable.
	 	*/
		lpar |= DHV_7BITS | DHV_PARITY_ENABLE ;

	/*
	 * Set to either even or odd parity.
	 */
	if( (tp->t_flags & EVENP) )
		lpar |= DHV_EVEN_PARITY ;
	else
		lpar &= ~DHV_EVEN_PARITY ;
	
	/*
	 * 110 baud... Are there any ASR 33's really still out there????
	 */
	if( (tp->t_ispeed) == B110 )
		lpar |= DHV_2STOP ;

	/* Set the register */
	dhv_addr->lpr = lpar ;
# ifdef POLLED
	/*
	 * If we are polling, we can wait too long to process a
	 * ^S/^Q sequence.  So, we will let the DHV do it in hard-
	 * ware for us.  RAW mode means give us everything, so I
	 * will let it pass ^S/^Q's.
	 */
	if( tp->t_flags & RAW ) {
	 	dhv_addr->lnctrl &= ~DHV_OAUTO ;	/* turn off hw flow */
		DELAY(1) ;				/* wait 20 us */
		dhv_addr->unn.tbuffad2 |= DHV_TX_ENA ;	
	}
	else {
		dhv_addr->unn.tbuffad2 |= DHV_TX_ENA ;
		/*
	 	 * I should check tchars to make sure that they are ^S/^Q.  On
	 	 * a DHV you can't change the definition.  I will leave this for
	 	 * the next revision.
	 	 */
		dhv_addr->lnctrl |= DHV_OAUTO ;		/* turn on hw flow */
	}
# endif

	(void)splx(s) ;
}

/*
 * DHV11 transmitter interrupt.  Restart each line which used to be active
 * but has terminated transmission since the last interrupt.
 */
dvxint( dhv )
int	dhv ;
{
register struct qb_device	*qi ;
register struct dhv_device	*dhv_addr ;
register int			unit ;
register struct tty		*tp ;
int				cntr ;
short				csr ;

	qi = dhv_info[dhv] ;
	dhv_addr = (struct dhv_device *)((qi->qi_mi)->qm_addr) ;
	while( (csr=dhv_addr->csr) & DHV_TX_ACTION ) {
		unit = ((csr>>8) & 0x7) | ( dhv << 3) ;
# ifdef POLLED
		dhv_addr->csr = (unit & 0x7) | DHV_TXIE ;
# else
		dhv_addr->csr = (unit & 0x7) | DHV_IE ;
# endif
		tp = &dhv_tty[unit] ;

		/*
		 * Check for DMA error.
		 */
		if( csr & DHV_TX_DMA_ERROR ) {
			(void)printf( "dhv%d: Xmit DMA error.\n", dhv ) ;
			dhv_addr->tbuffct = 0 ;
		}

		/* Xmit interrupt means we have completed a DMA transfer. */
		tp->t_state &= ~TS_BUSY ;

		/* If we are flushing the buffer... toss everything in the
		 * DMA buffer.
		 */
		if( tp->t_state & TS_FLUSH ) {
			tp->t_state &= ~TS_FLUSH ;
			dhv_addr->tbuffct = 0 ;
		} else {
			/*
			 * Compute how many bytes we sent out and remove them
			 * from the output queue.  We do calulations in 16 bit
			 * math.
			 */
			cntr = dhv_addr->tbuffad1 - (u_short)dhv_buf[unit] ;
			(void)ndflush( &tp->t_outq, (int)cntr) ;
		}

		/* Get ready to start another transfer */
		if( tp->t_line )
			(void)(*linesw[tp->t_line].l_start) (tp) ;
		else
			(void)dhv_start(tp) ;
	}

	return ;
}

/*
 * Start (restart) transmission on the given DHV11 line.
 */
dhv_start(tp)
register struct tty	*tp ;
{
register int			unit ;
register int			dhv ;
register struct dhv_device	*dhv_addr ;
register int			car ;
register int			nch ;
int				s ;

	unit = minor(tp->t_dev) ;
	dhv = dhv_board( unit ) ;
	dhv_addr = (struct dhv_device *)tp->t_addr ;
	/*
	 * Must hold interrupts in following code to prevent state of
	 * the tp from changing.
	 */
	s = spl5() ;
	
	/*
	 * if it's currently active, or delaying, no need to do anything.
	 */
	if (tp->t_state & (TS_TIMEOUT|TS_BUSY|TS_TTSTOP) ) {
		(void)splx(s) ;
		return ;
	}
	
	/*
	 * If there are any sleepers, and output has drianed below the low
	 * water mark, wake em up.
	 */
	if (tp->t_outq.c_cc <= TTLOWAT(tp) ) {
		if( tp->t_state & TS_ASLEEP ) {
			tp->t_state &= ~TS_ASLEEP ;
			(void)wakeup( (caddr_t)&tp->t_outq ) ;
		}
		if( tp->t_wsel ) {
			(void)selwakeup(tp->t_wsel, tp->t_state & TS_WCOLL );
			tp->t_wsel = 0 ;
			tp->t_state &= ~TS_WCOLL ;
		}
	}

	/*
	 * Now restart the transmission unless the output queue is empty.
	 */
	if (tp->t_outq.c_cc == 0 ) {
		(void)splx(s) ;
		return ;
	}

# ifdef POLLED
	dhv_addr->csr = ( unit & 0x7 ) | DHV_TXIE ;
# else
	dhv_addr->csr = ( unit & 0x7 ) | DHV_IE ;
# endif

	/*
	 * Transmit abort can be set several ways.  On is ^S/^Q hand
	 * shaking, another is if we want to Flush the buffer.  We
	 * abort output by setting TX_DMA_ABORT bit in dhv_stop.  We
	 * want to continue outputing on this channel, so clear it.
	 */
	if( dhv_addr->lnctrl & DHV_TX_DMA_ABORT )
		dhv_addr->lnctrl &= ~DHV_TX_DMA_ABORT ;
	
	/*
	 * Find out how many characters are in the output queue.
	 */
	if( tp->t_flags & (RAW|LITOUT) )
		nch = ndqb( &tp->t_outq, 0 ) ;
	else {
		nch = ndqb( &tp->t_outq, 0x80 ) ;
		/*
		 * If the first thing on the queue is a delay, process it.
		 */
		if (nch == 0 ) {
			nch = getc(&tp->t_outq) ;
			(void)timeout(ttrstrt, (caddr_t)tp, (nch&0x7f)+6) ;
			tp->t_state |= TS_TIMEOUT ;
			(void)splx(s) ;
			return ;
		}
	}

	/*
	 * If there is something to transmit, restart transmission.
	 */
	if (nch) {
		car = (int) dhv_buf[unit] ;
		/*
		 * copy the stuff out of the output queue to the buffer
		 * where we will actually do the DMA transfer from.
		 */
		(void)bcopy( tp->t_outq.c_cf, car, nch ) ;
		(void)byterev( car, nch ) ; /* for 68000 machines */
		/*
		 * Set up the address of the DMA buffer
		 */
		dhv_addr->tbuffad1 = car & 0xffff ;
		/*
		 * Tell the DHV how many characters the buffer holds.
		 */
		dhv_addr->tbuffct = nch ;
		/*
		 * start the DMA transfer.  Be careful not to mess with the
		 * high byte of the address buffer.  When we are in OAUTO
		 * mode we can't touch TX_ENA bit; otherwise, we could lose
		 * a ^S/^Q sequence.
		 *
		 * We also set the high bits of the buffer address.
		 */
		dhv_addr->unn.tbuffad2_ls[TB_LS] = ((car>>16) &  DHV_TBUFFAD2)
						 | DHV_DMA_START ;
		/*
		 * Set the DMA transfer busy bit in the users' tty structure.
		 */
		tp->t_state |= TS_BUSY ;
	}

	(void)splx(s) ;
	return ;
}

/*
 * Stop output on a line for ^S/^Q or output flush.
 */
/*ARGSUSED*/
dhv_stop(tp, flag)
register struct tty	*tp ;
{
register struct dhv_device	*dhv_addr  ;
register int			unit ;
int				s ;

	unit = minor(tp->t_dev) ;
	dhv_addr = (struct dhv_device *)tp->t_addr ;
	/*
	 * Block I/O interrupts while messing with state.
	 */
	s = spl5() ;
	if( tp->t_state & TS_BUSY) {
		/*
		 * Device is transmitting, stop the output by selectint the
		 * line and the setting the transmit abort.  We will get a
		 * transmit interrupt, where we will figure out where to
		 * continue the next time the trasmitter is enabled.
		 * dhv_start will clear the abort bit.
		 */
# ifdef POLLED
		dhv_addr->csr = (unit & 0x7) | DHV_TXIE ;
# else
		dhv_addr->csr = (unit & 0x7) | DHV_IE ;
# endif
		/*
		 * Check if we are flushing the buffer or if we are
		 * just ^S/^Qing.  Set the flush bit in the tty structure,
		 * if that is what we want to do.
		 */
		if ( (tp->t_state & TS_TTSTOP) == 0 )
			tp->t_state |= TS_FLUSH ;
		/*
		 * Abort the current transfer soonest.  dhv_start will reset
		 * this.  After dhvxint clears the BUSY bit in the users' tty
		 * structure.
		 */
		dhv_addr->lnctrl |= DHV_TX_DMA_ABORT ;
	}
	(void)splx(s) ;
}

/*
 * dhv_reset resets the state of the driver.
 * Reset the csr and lpr registers on open lines, and restart tranmitters.
 */
dhv_reset(qban)
int	qban ;
{
register	int		unit ;
register	int		dhv ;
register struct tty		*tp ;
register struct qb_device	*qi ;
register struct dhv_device	*dhv_addr ;
int				i ;

	for( dhv=0 ; dhv < NDHV; dhv++ ) {
		qi = dhv_info[dhv] ;
		if(qi==0 || qi->qi_alive==0)
			continue ;
		(void)printf( "dhv%d: reseting\n", dhv ) ;
		dhv_addr = (struct dhv_device *)qi->qi_mi->qm_addr ;
		unit = dhv * DHV_LINES ;
		for( i=0; i < DHV_LINES; i++ ) {
			tp = (struct tty *) &dhv_tty[unit] ;
# ifdef POLLED
			dhv_addr->csr = (unit & 0x7) | DHV_TXIE ;
# else
			dhv_addr->csr = (unit & 0x7) | DHV_IE ;
# endif
			if(tp->t_state & (TS_ISOPEN|TS_WOPEN)) {
				(void)dhv_param(unit) ;
				dhv_addr->lnctrl = DHV_ON | DHV_LINK_TYPE ;
				dhv_addr->unn.tbuffad2 |= DHV_TX_ENA ;
# ifdef POLLED
				if((tp->t_flags&RAW) == 0)
					dhv_addr->lnctrl |= DHV_OAUTO ;
# endif
				tp->t_state &= ~TS_BUSY ;
				(void)dhv_start(tp) ;
			}
			unit++ ;
		}
	}
# ifdef POLLED
	dhv_timer() ;
# endif
}

# ifdef POLLED
/*
 * At a software clock interrupt time empty all the dhv silos.
 * We will use a linear function to determine the polling rate.  I am
 * going to try:
 *		(Max - dhv_chars[dhv]) / 32 + 1
 *	Where:
 *		Max	  = The maximum characters that RBUF can hold.
 *		dhv_chars = The number of characters in the queue last time
 *			    around.
 * 
 * This function should break it up into 8 equal parts.  Where the Maximum
 * polling rate will be 1/60 of a second.  This allows
 *	256 * 60 characters / second = 15360 characters / second.  The maximum
 * a DHV can ever handle is 15000 characters / second.
 *
 * The slowest poll is 9/60 of a second.  This allows
 *	256 * 60/9 characters / second = 1706 characters / second.  Or about
 * 19.2k baud.
 *
 * Note: This algorithm will fail if there are a lot of very quick changes
 * on the line.  You will get silo overflows.  If that happens often you
 * should probably change the TIME_FUNCTION to something that better
 * predicts what happens on serial lines or just turn interrupts back on
 * again.
 *
 * The reason we do this is to avoid interrupting for every character that
 * comes in.  On a DH we really didn't have to worry about this because you
 * could set the silos to interrupt after it received 16 characters.
 */
# define	TIME_FUNCTION(x) ((( 256 - ((x>256)? 256:x) ) >> 5) + 1)
dhv_timer()
{
register int dhv ;
register int s ;
register int max_chars ;

	max_chars = 0 ;
	s=spl5() ;
	for( dhv=0; dhv < NDHV; dhv++ ) {
		(void)dvrint( dhv) ;
		if( dhv_chars[dhv] > max_chars ) {
			max_chars = dhv_chars[dhv] ;
			dhv_chars[dhv] = 0 ;
		}
	}
	(void)splx(s) ;
	(void)timeout( dhv_timer, (caddr_t)0, TIME_FUNCTION( max_chars ) ) ;
	return ;
}
# endif POLLED 
#endif DHV

