#include "../is68kif/if_ring.h"

#define	DELAY_COUNT	5000	/* how high to count when waiting on TAS */

/*
 * Initialize the ring pointed to by "rp" to contain "size" slots. The ring 
 * is set up to be empty and unlocked.  Size had better be a power of 2!
 */
vbringinit(rp, size)
	register struct ring *rp;
	short size;
{
	register int i;

	rp->r_tas = rp->r_rdidx = rp->r_wrtidx = 0;
	rp->r_size = size;
	for (i=0; i < size; i++)
		rp->r_slot[i] = 0;
}

/*
 * Add a new entry "v" to the ring "rp".  Returns 0 if ring was full, 
 * else the new number of entries in the ring.  Uses test-and-set routine, 
 * allowing multi-processor interlocked access to the ring IF the ring is 
 * in VME memory. Masks all interrupts, in order to prevent reentrancy and to
 *  minimize ring lockout.
 */
vbringput(rp, v)
	register struct ring *rp;
	int v;
{
	register int x = DELAY_COUNT;	/* don't wait forever for lock */
	register int s;

	for (;;) {	
		s = splhigh();		/* disable interrupts */ 
		if (tas(&rp->r_tas))	/* lock the ring */
			break;
		splx(s);		/* someone else has it */
		if (--x == 0)
			return 0;	/* ring is stuck! */
	}
	if ((x = (rp->r_wrtidx + 1) & (rp->r_size - 1)) != rp->r_rdidx) {
		rp->r_slot[rp->r_wrtidx] = v;
		rp->r_wrtidx = x;
	}				/* if x == rdidx then ring is full */
	if ((x -= rp->r_rdidx) < 0)
		x += rp->r_size;
	rp->r_tas = 0;			/* unlock ring */
	splx(s);
	return x;
}

/*
 * Extract next entry from the ring "rp". Returns 0 if ring is empty, else 
 * the extracted entry. Uses test-and-set routine, allowing multi-processor 
 * interlocked access to the ring IF the ring is in VME memory.  Masks all 
 * interrupts, in order to prevent reentrancy and to minimize ring lockout.
 */
vbringget(rp)
	register struct ring *rp;
{
	register int x = DELAY_COUNT;	/* don't wait forever for lock */
	register int s;

	for (;;) {
		s = splhigh();		/* disable interrupts */ 
		if (tas(&rp->r_tas))	/* lock the ring */
			break;
		splx(s);		/* someone else has it */
		if (--x == 0)
			return 0;	/* ring is stuck! */
	}

	if (rp->r_rdidx != rp->r_wrtidx) {
		x = rp->r_slot[rp->r_rdidx];
		rp->r_slot[rp->r_rdidx] = 0;
		rp->r_rdidx = (rp->r_rdidx + 1) & (rp->r_size - 1);
	} else
		x = 0;
	rp->r_tas = 0;			/* unlock ring */
	splx(s);			/* restore previous ipl */
	return x;
}

vbringempty(rp)
	register struct ring *rp;
{
	return rp->r_rdidx == rp->r_wrtidx;
}
